import numpy as np
import os
from PIL import Image

from length.constants import DTYPE
from length.data_sets import MNIST


def test_mnist_with_scale():
    with Image.open(os.path.join("length", "tests", "data", "5.png")) as image:
        scale = 2.0
        expected = np.array(image) * scale / 255
        data_set = MNIST(1, sample_dimensions=2, scale=scale)
        for batch in data_set.train:
            first_sample, first_label = batch.data.data[0], batch.labels.data[0]
            np.testing.assert_almost_equal(expected, first_sample)
            assert first_label == 5
            break


def test_mnist_without_scale():
    with Image.open(os.path.join("length", "tests", "data", "5.png")) as image:
        expected = np.array(image)
        data_set = MNIST(1, sample_dimensions=2, scale=None)
        for batch in data_set.train:
            first_sample, first_label = batch.data.data[0], batch.labels.data[0]
            np.testing.assert_equal(expected, first_sample)
            assert first_label == 5
            break


def test_mnist_reading_test():
    data_set = MNIST(100)
    iterations = 0
    for batch in data_set.train:
        assert (100, 784) == batch.data.shape
        assert DTYPE == batch.data.dtype
        assert (100,) == batch.labels.shape
        assert np.uint8 == batch.labels.dtype
        iterations += 1
    assert iterations == 600


def test_mnist_reading_train():
    data_set = MNIST(100)
    iterations = 0
    for batch in data_set.test:
        assert (100, 784) == batch.data.shape
        assert DTYPE == batch.data.dtype
        assert (100,) == batch.labels.shape
        assert np.uint8 == batch.labels.dtype
        iterations += 1
    assert iterations == 100


def test_mnist_reading_2_dimensions():
    data_set = MNIST(100, 2)
    iterations = 0
    for batch in data_set.train:
        assert (100, 28, 28) == batch.data.shape
        assert DTYPE == batch.data.dtype
        assert (100,) == batch.labels.shape
        assert np.uint8 == batch.labels.dtype
        iterations += 1
    assert iterations == 600


def test_mnist_reading_3_dimensions():
    data_set = MNIST(100, 3)
    iterations = 0
    for batch in data_set.train:
        assert (100, 1, 28, 28) == batch.data.shape
        assert DTYPE == batch.data.dtype
        assert (100,) == batch.labels.shape
        assert np.uint8 == batch.labels.dtype
        iterations += 1
    assert iterations == 600



import numpy as np

from length.function import Function


class Relu(Function):
    """
    The Relu Layer is a non-linear activation
    """
    name = "ReLU"

    def __init__(self):
        super().__init__()
        # TODO: add more initialization if necessary

    def internal_forward(self, inputs):
        x, = inputs
        # TODO: calculate forward pass of ReLU function
        return x,

    def internal_backward(self, inputs, gradients):
        x, = inputs
        grad_in, = gradients
        # TODO: calculate gradients of ReLU function with respect to the input
        return grad_in,


def relu(x):
    """
    This function computes the element-wise ReLU activation function (https://en.wikipedia.org/wiki/Rectifier_(neural_networks))
    on a given input vector x.
    :param x: the input vector
    :return: a rectified version of the input vector
    """
    return Relu()(x)



import numpy as np

from length.functions.add import add, Add
from length.graph import Graph
from length.tests import gradient_checker
from length.tests.utils import init, retry


def test_add_forward():
    data = init([2])

    result = add(Graph(data), Graph(data))
    assert result.data == 4


@retry(3)
def test_add_backward():
    data = init([2])
    gradient = init([1])

    add_function = Add()
    computed_gradients_1, computed_gradients_2 = add_function.backward(gradient)

    data_copy = np.copy(data)
    f = lambda: add_function.internal_forward((data, data_copy))
    numerical_gradients_1, numerical_gradients_2 = gradient_checker.compute_numerical_gradient(f, (data, data_copy), (gradient,))

    gradient_checker.assert_allclose(computed_gradients_1, numerical_gradients_1)
    gradient_checker.assert_allclose(computed_gradients_2, numerical_gradients_2)



from length.function import Function


class Layer(Function):
    """
    Abstract Layer is a super class for all neural network layers. A layer behaves like a function, but also
    has to keep track of internal parameters (like weights and biases) that need to be optimized.
    """
    needs_optimizer = True
    name = "Layer"

    def __init__(self):
        super().__init__()
        self.optimizer = None

    def internal_update(self, parameter_deltas):
        """
        :param parameter_deltas: contains the delta of the parameters to be applied to each parameter
        (same structure as in internal_backward but without the first element)
        """
        raise NotImplementedError

    def backward(self, gradients):
        gradients = super().backward(gradients)
        input_gradient = gradients[:len(self.inputs)]
        parameter_gradients = gradients[len(self.inputs):]
        if len(parameter_gradients) > 0:
            parameter_deltas = self.optimizer.run_update_rule(parameter_gradients, self)
            self.internal_update(parameter_deltas)
        return input_gradient



import numpy as np

from length import constants
from length.function import Function


class MeanSquaredError(Function):
    """
    This function calculates the Mean Squared Error between two given vectors, as described in
    https://en.wikipedia.org/wiki/Mean_squared_error
    """
    name = "MeanSquaredError"

    def __init__(self):
        super().__init__()
        self.difference = None

    @staticmethod
    def create_one_hot(data, shape):
        assert len(shape) == 2, "Providing integers as second input to MSE only works with two dimensional input vectors"
        data_container = np.zeros(shape, dtype=constants.DTYPE)
        data_container[np.arange(len(data)), data] = 1
        return data_container

    def internal_forward(self, inputs):
        x1, x2 = inputs

        if np.issubdtype(x2.dtype, np.integer):
            x2 = self.create_one_hot(x2, x1.shape)

        self.difference = x1 - x2
        squared_sum = np.sum(np.square(self.difference))
        return (squared_sum / self.difference.size).astype(constants.DTYPE),

    def internal_backward(self, inputs, gradients):
        x1, x2 = inputs
        gx, = gradients
        derived_value = 2 / x1.size * self.difference
        gradient = derived_value * gx

        if np.issubdtype(x2.dtype, np.integer):
            # in case we used MSE as loss function, we won't propagate any gradients to the loss
            return gradient, None

        return gradient, -gradient


def mean_squared_error(input_1, input_2):
    """
    This function calculates the Mean Squared Error between input_1 and input_2. Both inputs should be vectors of the
    same shape. You can also supply a one-dimensional list of integers.
    If you do so this vector will be converted to a one_hot representation that fits to the shape of the second
    input
    :param input_1: the first vector of any shape
    :param input_2: the second vector. Needs to have the same shape as the first vector, or be a one-dimensional int vector
    :return: the mean squared error between input_1 and input_2
    """
    return MeanSquaredError()(input_1, input_2)


import numpy as np

from length.function import Function
from length.functions.softmax import log_softmax


class SoftmaxCrossEntropy(Function):
    name = "SoftmaxCrossEntropy"

    def __init__(self):
        super().__init__()
        self.y = None

    def internal_forward(self, inputs):
        x, t = inputs
        self.y = log_softmax(x)
        loss = -self.y[range(len(t)), t].sum(keepdims=True) / t.size
        return loss[0],

    def internal_backward(self, inputs, gradients):
        x, t = inputs
        grad_in, = gradients
        grad_x = np.exp(self.y.copy())
        grad_x[range(len(t)), t] -= 1
        grad_x *= grad_in[0] / t.size
        return grad_x, None


def softmax_cross_entropy(x, t):
    """
    This function calculates the softmax cross entropy loss
    (https://cs231n.github.io/linear-classify/#softmax-classifier)between a given two-dimensional input vector `x` and a
    one-dimensional int-vector `t` that represents the groundtruth labels for classification.
    :param x: the input vector that is to be used to calculate the loss
    :param t: the groundtruth labels for the batch
    :return: the cross entropy loss of input and groundtruth labels
    """
    return SoftmaxCrossEntropy()(x, t)



import numpy as np

from length import constants
from length.functions.softmax import Softmax, softmax
from length.graph import Graph
from length.tests import gradient_checker
from length.tests.utils import retry


def test_softmax_forward():
    data = np.random.uniform(-1, 1, (3, 10)).astype(constants.DTYPE)

    output = softmax(Graph(data)).data

    expected_output = np.exp(data)
    for i in range(len(output)):
        expected_output[i] /= expected_output[i].sum()

    gradient_checker.assert_allclose(output, expected_output)


@retry(3)
def test_softmax_backward():
    data = np.random.uniform(-1, 1, (3, 10)).astype(constants.DTYPE)
    gradient = np.random.uniform(-1, 1, (3, 10)).astype(constants.DTYPE)

    data_graph = Graph(data)
    softmax_function = Softmax()
    softmax_function(data_graph)
    computed_gradients, = softmax_function.backward(gradient)

    f = lambda: softmax_function.internal_forward((data,))
    numerical_gradients, = gradient_checker.compute_numerical_gradient(f, (data,), (gradient,), eps=1e-2)

    gradient_checker.assert_allclose(computed_gradients, numerical_gradients)



import numpy as np

from length.optimizer import Optimizer


class Adam(Optimizer):
    """
    The Adam optimizer (see https://arxiv.org/abs/1412.6980)
    :param learning_rate: initial step size
    :param beta1: Exponential decay rate of the first order moment
    :param beta2: Exponential decay rate of the second order moment
    :param eps: Small value for numerical stability
    """

    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, eps=1e-8):
        self.learning_rate = learning_rate
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps

        # TODO: add more initialization code

    def run_update_rule(self, gradients, layer):
        # TODO: implement Adam update rule as specified in https://arxiv.org/abs/1412.6980
        param_deltas = [np.zeros_like(g) for g in gradients]
        return param_deltas



import numpy as np

from length.data_sets import MNIST, FashionMNIST
from length.models import MLP
from length.optimizers import Adam, SGD


def test_train():
    np.seterr(divide='raise')

    for optimizer in [SGD(0.001), Adam(0.001)]:
        for data_set in [MNIST(64), FashionMNIST(64)]:
            for i in range(10):
                model = MLP()

                for iteration, batch in enumerate(data_set.train):
                    model.forward(batch)
                    model.backward(optimizer)
                    break



import numpy as np

from length.function import Function


class Dropout(Function):
    name = "Dropout"

    def __init__(self, dropout_ratio):
        super().__init__()
        if not 0.0 <= dropout_ratio < 1:
            raise ValueError("dropout_ratio must be in range [0, 1)")
        self.dropout_ratio = dropout_ratio
        # TODO: add more initialization if necessary

    def internal_forward(self, inputs):
        x, = inputs
        # TODO: implement forward pass of dropout function
        return x,

    def internal_backward(self, inputs, gradients):
        gradient, = gradients
        # TODO: implement backward pass of dropout function
        return gradient,


def dropout(x, dropout_ratio=0.5, train=True):
    """
    This function implements dropout (http://www.jmlr.org/papers/volume15/srivastava14a/srivastava14a.pdf),
    a regularization method for neural networks.
    :param x: the input vector where parts shall be dropped
    :param dropout_ratio: the ratio of which to perform dropout
    :param train: whether we are currently running in train or testing mode (default: True)
    :return: a vector with a portion of elements zeroed out, this portion is defined by `dropout_ratio`.
    """
    # TODO: call the dropout function and handle the scenario if we are not training
    return x



from length.optimizer import Optimizer


class SGD(Optimizer):
    """
    An optimizer that does plain Stochastic Gradient Descent
    (https://en.wikipedia.org/wiki/Stochastic_gradient_descent#Iterative_method)
    :param lr: the learning rate to use for optimization
    """

    def __init__(self, lr):
        self.lr = lr

    def run_update_rule(self, gradients, _):
        # TODO: implement SGD update rule
        # HINT: it should not be more than one line ;)
        param_deltas = None
        return param_deltas



import math

import numpy as np

from length.functions import mean_squared_error
from length.functions.mean_squared_error import MeanSquaredError
from length.graph import Graph
from length.tests import gradient_checker
from length.tests.utils import init, retry


def fixed_case(with_label=False):
    data = init([
        [0, 0, 0, 1],
        [1, 0, 0, 1],
        [2, 0, 1, 0],
    ])

    if with_label:
        data_2 = np.array([3, 0, 1], dtype=np.int32)
    else:
        data_2 = init([
            [1, 0, 0, 1],
            [0, 0, 0, 1],
            [0, 0, 0, 0]
        ])

    return data, data_2


def test_mean_squared_error_forward_zero_loss():
    data = Graph(init([0, 0, 0, 1]))
    label = Graph(init([0, 0, 0, 1]))

    mse = mean_squared_error(data, label)

    assert float(mse.data) == 0


def test_mean_squared_error_forward_loss():
    data, data_2 = fixed_case()

    mse = mean_squared_error(Graph(data), Graph(data_2))
    assert math.isclose(float(mse.data), 0.583, abs_tol=1e-3)


def test_mean_squared_error_forward_int_input():
    data, labels = fixed_case(with_label=True)

    mse = mean_squared_error(Graph(data), Graph(labels))
    assert math.isclose(float(mse.data), 0.583, abs_tol=1e-3)


@retry(3)
def test_mean_squared_error_backward():
    data, data_2 = fixed_case()
    gradients = init([2])

    data_1_graph = Graph(data)
    data_2_graph = Graph(data_2)

    mse_function = MeanSquaredError()
    mse_function(data_1_graph, data_2_graph)
    computed_gradient_1, computed_gradient_2 = mse_function.backward(gradients)

    f = lambda: mse_function.internal_forward((data, data_2))
    numerical_gradient_1, numerical_gradient_2 = gradient_checker.compute_numerical_gradient(
        f,
        (data, data_2),
        (gradients,)
    )

    gradient_checker.assert_allclose(computed_gradient_1, numerical_gradient_1)
    gradient_checker.assert_allclose(computed_gradient_2, numerical_gradient_2)


@retry(3)
def test_mean_squared_error_backward_with_label():
    data, data_2 = fixed_case(with_label=True)
    gradients = init([2])

    data_1_graph = Graph(data)
    data_2_graph = Graph(data_2)

    mse_function = MeanSquaredError()
    mse_function(data_1_graph, data_2_graph)
    computed_gradient_1, computed_gradient_2 = mse_function.backward(gradients)
    assert computed_gradient_2 is None

    f = lambda: mse_function.internal_forward((data, data_2))
    numerical_gradient_1, _ = gradient_checker.compute_numerical_gradient(
        f,
        (data, data_2),
        (gradients,)
    )

    gradient_checker.assert_allclose(computed_gradient_1, numerical_gradient_1)



import math

import numpy as np
import pytest

from length import constants
from length.functions import dropout
from length.functions.dropout import Dropout
from length.graph import Graph
from length.tests import gradient_checker
from length.tests.utils import retry


def _dropout(input, function):
    return input * function.mask


def get_data():
    data = np.random.random((4, 5)).astype(dtype=constants.DTYPE)
    return data


def test_dropout_forward_ratio_0():
    data = get_data()

    result = dropout(Graph(data), dropout_ratio=0.)
    gradient_checker.assert_allclose(result.data, data)


def test_dropout_forward_ratio_1():
    data = get_data()

    with pytest.raises(ValueError):
        dropout(Graph(data), dropout_ratio=1.)


def test_dropout_forward_ratio_0_5():
    data = get_data()

    # set the seed to get reliable results
    np.random.seed(2)

    result = dropout(Graph(data), dropout_ratio=0.5)
    # check that ca. 50% of all data points are zero now
    non_zero_elements = result.data.nonzero()
    assert math.isclose(data.size // 2, non_zero_elements[0].size, abs_tol=data.size * 0.1)

    # reset the seed in case this might matter
    np.random.seed()


@retry(3)
def test_dropout_backward():
    data = get_data()
    gradient = np.random.random(data.shape).astype(constants.DTYPE)

    data_graph = Graph(data)
    dropout_function = Dropout(0.5)
    dropout_result = dropout_function(data_graph)
    computed_gradients, = dropout_function.backward(gradient)

    f = lambda: _dropout(data, dropout_result.creator)
    numerical_gradients, = gradient_checker.compute_numerical_gradient(f, (data,), gradient, eps=0.1)

    gradient_checker.assert_allclose(computed_gradients, numerical_gradients)


def test_ropout_test_mode():
    data = get_data()

    dropout_result = dropout(data, dropout_ratio=0.5, train=False)
    # there should be no changes in test mode
    gradient_checker.assert_allclose(dropout_result.data, data)



import numpy as np

from length import constants
from length.functions.sum import Sum, sum
from length.graph import Graph
from length.tests import gradient_checker
from length.tests.utils import init, retry


def test_sum_forward():
    data = np.random.uniform(-1, 1, (3, 2)).astype(constants.DTYPE)

    sum_output = sum(Graph(data))
    np.testing.assert_allclose(sum_output.data, data.sum())


@retry(3)
def test_sum_backward():
    data = np.random.uniform(-1, 1, (3, 2)).astype(constants.DTYPE)
    gradient = init([2])

    data_graph = Graph(data)
    sum_function = Sum()
    sum_function(data_graph)
    computed_gradients, = sum_function.backward((gradient,))

    f = lambda: sum_function.internal_forward((data,))
    numerical_gradients, = gradient_checker.compute_numerical_gradient(f, (data,), (gradient,))

    gradient_checker.assert_allclose(computed_gradients, numerical_gradients, atol=1e-4, rtol=1e-3)



import numpy as np

from length.data_set import Batch
from length.models import MLP


def test_graph_backward_no_layers():
    model = MLP()
    model.forward(Batch(
        np.random.random((10, 784)),
        np.random.randint(0, 10, 10)
    ))
    string = model.loss._visualization_as_str()

    assert "id" in string
    assert "layer" in string
    assert "next" in string

    def assert_all_included(layer):
        if layer.creator is None:
            return
        assert repr(layer) in string
        for predecessor in layer.predecessors:
            assert_all_included(predecessor)

    assert_all_included(model.loss)

    column_lengths = [[len(column) for column in line.split("|")] for line in string.split("\n")]
    first = column_lengths[0]
    for other in column_lengths[1:]:
        for x, y in zip(first, other):
            assert x == y



from length.layer import Layer
from length.layers import *


def test_layer_needs_optimizer():
    # here we test that each layers wants to have an optimizer
    for sub_class in Layer.__subclasses__():
        assert sub_class.needs_optimizer is True





class Optimizer:
    """
    An optimizer applies a certain update rule on given gradients and returns the parameter deltas
    """

    def run_update_rule(self, gradients, layer):
        """
        Does the actual optimization step and calculates the update
        :param gradients: the gradients of the layer that are to be used to optimize the parameters (a tuple of
        numpy arrays)
        :param layer: the layer to which the gradients belong
        :return: the deltas that are to be applied to the parameters
        """
        raise NotImplementedError



import numpy as np

from length.function import Function


def log_softmax(x):
    def logsumexp(x):
        m = x.max(axis=1, keepdims=True)
        y = x - m
        np.exp(y, out=y)
        s = y.sum(axis=1, keepdims=True)
        np.log(s, out=s)
        m += s
        return m

    log_z = logsumexp(x)
    # log(e^(x)/log_z)
    y = x - log_z
    return y


class Softmax(Function):
    name = "Softmax"

    def __init__(self):
        super().__init__()
        self.y = None

    def internal_forward(self, inputs):
        x, = inputs
        assert x.ndim == 2, "Softmax only supports two-dimensional input"
        self.y = np.exp(log_softmax(x))
        return self.y,

    def internal_backward(self, inputs, gradients):
        x, = inputs
        grad_in, = gradients
        grad_x = self.y * grad_in
        sum_grad_x = grad_x.sum(axis=1, keepdims=True)
        grad_x -= self.y * sum_grad_x

        assert x.shape == grad_x.shape

        return grad_x,


def softmax(x):
    """
    The softmax function takes a two-dimensional input (of shape BxN) and computes the softmax function
    (https://en.wikipedia.org/wiki/Softmax_function)on the last dimension of this input. The result is a two-dimensional
    vector (of shape BxN) that contains a probability distribution for each sample b of the batch with batch size B.
    :param x: the two-dimensional input vector to calculate the softmax on
    :return: a two-dimensional vector with softmax applied to the last dimension
    """
    return Softmax()(x)



from length.layers import FullyConnected
import length.functions as F


class MLP:
    """
    Simple network with two fully connected layers
    """
    def __init__(self):
        self.fully_connected_1 = FullyConnected(784, 512)
        self.fully_connected_2 = FullyConnected(512, 512)
        self.fully_connected_3 = FullyConnected(512, 10)

        self.loss = None
        self.predictions = None

    def forward(self, batch, train=True):
        """
        This runs the forward pass with the model
        :param batch: the batch which should be forwarded
        """
        hidden = self.fully_connected_1(batch.data)
        hidden = self.fully_connected_2(hidden)
        self.predictions = self.fully_connected_3(hidden)
        self.loss = F.mean_squared_error(self.predictions, batch.labels)

    def backward(self, optimizer):
        """
        This runs the backward pass (should *not* be used in test loop)
        :param optimizer: the optimizer which calculated the updates
        """
        self.loss.backward(optimizer)



import numpy as np

from length.function import Function


class Sum(Function):
    name = "Sum"

    def __init__(self):
        super().__init__()

    def internal_forward(self, inputs):
        x, = inputs
        return np.sum(x),

    def internal_backward(self, inputs, gradients):
        x, = inputs
        grad_in, = gradients
        grad_x = np.full_like(x, grad_in)
        assert grad_x.shape == x.shape
        return grad_x,


def sum(x):
    """
    This function calculates the sum of all elements in a given vector
    :param x: the vector that is to be summed.
    :return: the sum of all elements in the input vector
    """
    return Sum()(x)



import os
import shutil
import pytest

from length.data_sets import MNIST, FashionMNIST

temp_folder = ".temp"


@pytest.fixture(scope="module")
def cleanup():
    yield cleanup
    print("delete temp folder")
    if os.path.isdir(temp_folder):
        shutil.rmtree(temp_folder)


def test_mnist_downloading(cleanup):
    data_set = MNIST(10, delay_loading=True)
    data_set.path = temp_folder
    data_set.download_files()

    required_files = [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
    ]

    for file in required_files:
        assert os.path.isfile(os.path.join(temp_folder, file))


def test_fashion_mnist_downloading(cleanup):
    data_set = FashionMNIST(10, delay_loading=True)
    data_set.path = temp_folder
    data_set.download_files()

    required_files = [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
    ]

    for file in required_files:
        assert os.path.isfile(os.path.join(temp_folder, file))



import numpy as np
import os
import gzip
import struct
from urllib.request import urlretrieve

from length.constants import DTYPE
from length.data_set import DataSet, Batch


class MNISTLike(DataSet):
    """
    Class for all data sets which are in the same format as MNIST
    """
    name = None
    url = None

    def __init__(self, batch_size, sample_dimensions=1, load_train=True, load_test=True, delay_loading=False,
                 scale=1.0, **kwargs):
        super().__init__(batch_size, **kwargs)
        self.data_url = self.url
        self.path = os.path.join(".data", self.name)
        self.sample_dimensions = sample_dimensions
        self.scale = scale

        self.load_train = load_train
        self.load_test = load_test

        self.train_images = None
        self.train_labels = None
        self.test_images = None
        self.test_labels = None

        # store which file should be written to which property
        self.basenames = {}
        if self.load_train:
            self.basenames["train-images-idx3-ubyte.gz"] = "train_images"
            self.basenames["train-labels-idx1-ubyte.gz"] = "train_labels"
        if self.load_test:
            self.basenames["t10k-images-idx3-ubyte.gz"] = "test_images"
            self.basenames["t10k-labels-idx1-ubyte.gz"] = "test_labels"

        if not delay_loading:
            self.prepare()

    @property
    def test(self):
        return self.batch_iterator(self.test_images, self.test_labels)

    @property
    def train(self):
        return self.batch_iterator(self.train_images, self.train_labels)

    def batch_iterator(self, data_array, label_array):
        for i in range(len(label_array) // self.batch_size):
            data = data_array[self.batch_size * i:self.batch_size * (i + 1)]
            labels = label_array[self.batch_size * i:self.batch_size * (i + 1)]
            yield Batch(data, labels)

    def prepare(self):
        self.download_files()
        self.load_files_into_memory()
        self.check_sanity()

    def download_files(self):
        os.makedirs(self.path, exist_ok=True)

        urls = [self.data_url + f for f in self.basenames]
        for url, basename in zip(urls, self.basenames):
            output_file_name = os.path.join(self.path, basename)
            if not os.path.isfile(output_file_name):
                print("downloading file %s..." % basename)
                urlretrieve(url, output_file_name)

    def load_files_into_memory(self):
        for file_name, target in self.basenames.items():
            file_path = os.path.join(self.path, file_name)
            with gzip.open(file_path) as handle:
                # see IDX file format specification:
                # http://yann.lecun.com/exdb/mnist/
                dtypes = {
                    "08": np.uint8,
                    "09": np.int8,
                    "0B": np.int16,
                    "0C": np.int32,
                    "0D": np.float32,
                    "0E": np.float64,
                }

                magic_num = handle.read(4).hex()
                assert "0000" == magic_num[:4]
                dtype = dtypes[magic_num[4:6]]
                num_dim = int(magic_num[6:8], 16)

                dimensions = [int(handle.read(4).hex(), 16) for _ in range(num_dim)]

                data = np.fromstring(handle.read(), dtype=dtype)

                if "images" in target:
                    # only do this if we are reading an image file
                    shape = dimensions
                    if num_dim > 1 + self.sample_dimensions:
                        shape = [dimensions[0]] + [np.prod(dimensions[1:])]
                    elif num_dim < 1 + self.sample_dimensions:
                        shape = [dimensions[0]] + [1] + dimensions[1:]

                    data = data.reshape(shape)

                    if self.scale is not None:
                        # convert data to internally used dtype (float)
                        data = data.astype(DTYPE) / 255 * self.scale

                # set loaded data
                setattr(self, target, data)

    def check_sanity(self):
        if self.load_test:
            assert self.test_images is not None
            assert self.test_labels is not None
            assert len(self.test_labels) == len(self.test_images)

        if self.load_train:
            assert self.train_images is not None
            assert self.train_labels is not None
            assert len(self.train_labels) == len(self.train_images)


from .mnist_like import MNISTLike


class FashionMNIST(MNISTLike):
    """
    The replacement of the hand-written digit data set with fashion items
    """

    name = "fashion"
    url = "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/"



import numpy as np

from length.layer import Layer
from length.constants import DTYPE
from length.initializers.xavier import Xavier


class FullyConnected(Layer):
    """
    The FullyConnected Layer is one of the base building blocks of a neural network. It computes a weighted sum
    over the input, using a weight matrix. It furthermore applies a bias term to this weighted sum to allow linear
    shifts of the computed values.
    """
    name = "FullyConnected"

    def __init__(self, num_inputs, num_outputs, weight_init=Xavier()):
        super().__init__()

        self._weights = np.zeros((num_outputs, num_inputs,), dtype=DTYPE)
        weight_init(self._weights)
        self.bias = np.zeros((num_outputs,), dtype=DTYPE)

    @property
    def weights(self):
        return self._weights.T

    @weights.setter
    def weights(self, value):
        self._weights = value.T

    def internal_forward(self, inputs):
        x, = inputs
        return np.dot(x, self._weights.T) + self.bias,

    def internal_backward(self, inputs, gradients):
        x, = inputs
        grad_in, = gradients
        grad_x = np.dot(grad_in, self._weights)
        grad_w = np.dot(grad_in.T, x)
        grad_b = np.sum(grad_in, axis=0)

        assert grad_x.shape == x.shape
        assert grad_w.shape == self._weights.shape
        assert grad_b.shape == self.bias.shape

        return grad_x, grad_w, grad_b

    def internal_update(self, parameter_deltas):
        delta_w, delta_b = parameter_deltas
        self._weights -= delta_w
        self.bias -= delta_b


import numpy as np

from length.initializer import Initializer


class Xavier(Initializer):
    """
    Xavier initializer that initializes weights following the initialization scheme proposed in:
    http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf?hc_location=ufi
    """

    def __init__(self, scale=1.0):
        super().__init__()
        self.scale = scale

    def __call__(self, array):
        # TODO: overwrite values in array with a correctly initialized one
        array[...] = np.random.normal(loc=0.0, scale=1e-2, size=array.shape)



from length.graph import Graph


class Batch:
    def __init__(self, data, labels):
        self.data = Graph(data, name="input[data]")
        self.labels = Graph(labels, name="input[labels]")


class DataSet:
    """
    Abstract DataSet, it should implement ways to retrieve test and train data and support shuffling
    """
    def __init__(self, batch_size, shuffle=True, repeat=False):
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.repeat = repeat

    @property
    def test(self):
        """
        :return: an iterator over the test set.
        """
        raise NotImplementedError()

    @property
    def train(self):
        """
        :return: an iterator over the train set.
        """
        raise NotImplementedError()



from .accuracy import accuracy
from .add import add
from .dropout import dropout
from .mean_squared_error import mean_squared_error
from .relu import relu
from .sigmoid import sigmoid
from .softmax import softmax
from .softmax_cross_entropy import softmax_cross_entropy
from .sum import sum



import math

import numpy as np

from length import constants
from length.functions import softmax_cross_entropy
from length.functions.softmax_cross_entropy import SoftmaxCrossEntropy
from length.graph import Graph
from length.tests import gradient_checker
from length.tests.utils import init, retry


def get_data():
    data = np.random.uniform(-1, 1, (10, 20)).astype(constants.DTYPE)
    labels = np.random.randint(0, 20, (10,)).astype(np.int32)
    return data, labels


def test_softmax_cross_entropy_forward():
    data, labels = get_data()

    softmax_loss = softmax_cross_entropy(Graph(data), Graph(labels))

    y = np.exp(data)
    expected_loss = 0
    for i in range(len(y)):
        expected_loss -= math.log(y[i, labels[i]] / y[i].sum())
    expected_loss /= len(y)

    assert math.isclose(float(softmax_loss.data), expected_loss, rel_tol=1e-4, abs_tol=1e-5)


@retry(3)
def test_softmax_cross_entropy_backward():
    data, labels = get_data()
    gradient = init([2])

    loss_function = SoftmaxCrossEntropy()
    loss_function(Graph(data), Graph(labels))
    computed_gradient_data, computed_gradient_label = loss_function.backward(gradient)
    assert computed_gradient_label is None

    f = lambda: loss_function.internal_forward((data, labels))
    numerical_gradient_data, _ = gradient_checker.compute_numerical_gradient(f, (data, labels), (gradient,), eps=1e-2)

    gradient_checker.assert_allclose(computed_gradient_data, numerical_gradient_data, atol=1e-4)



import numpy as np

from length import constants


def init(array):
    return np.array(array, dtype=constants.DTYPE)


def retry(max_num_retries):
    def retry_wrapper(func):
        def wrapper():
            num_errored = 0
            error = None
            for _ in range(max_num_retries):
                try:
                    func()
                    break
                except Exception as e:
                    num_errored += 1
                    error = e

            if num_errored == max_num_retries:
                raise error

        return wrapper
    return retry_wrapper



import numpy as np

from length.function import Function
from length.functions.softmax import Softmax


class Accuracy(Function):
    name = "Accuracy"

    def __init__(self):
        super().__init__()
        self.softmax = Softmax()

    def internal_forward(self, inputs):
        x, t = inputs
        softmaxed_x, = self.softmax.internal_forward((x,))
        predicted_classes = np.argmax(softmaxed_x, axis=1)

        num_correct = (predicted_classes == t).sum()

        return num_correct / len(t),

    def internal_backward(self, inputs, gradients):
        return None, None


def accuracy(x, t):
    """
    The accuracy function takes the output of a classifier (i.e. the last fully connected layer) and also an int-vector
    with groundtruth labels and calculates the accuracy of this prediction. Please Note: This function can not be used
    for backpropagation!
    :param x: a two-dimensional (Shape: (B,N) with B being the batch_size) vector that represents the classification
    result of the network.
    :param t: a one-dimensional vector of ints that represents the groundtruth labels
    :return: the accuracy of the classification.
    """
    return Accuracy()(x, t)



import numpy as np

import length.functions as F
from length.graph import Graph
from length.tests.utils import init


def calc_accuracy(data, labels):
    accuracy = F.accuracy(Graph(data), Graph(labels))
    return accuracy


def get_base_data():
    first_element = init([-10, -10, -5, -4, 5, -1])
    second_element = init([-10, -10, 7, 3, 4, -1])
    data = np.stack((first_element, second_element))
    labels = np.array([4, 0])
    return data, labels


def test_accuracy_forward():
    data, labels = get_base_data()

    accuracy = calc_accuracy(data, labels)
    assert accuracy.data == 0.5

    labels = np.array([4, 2])
    accuracy = calc_accuracy(data, labels)
    assert accuracy.data == 1


def test_accuracy_backward():
    data, labels = [Graph(x) for x in get_base_data()]

    accuracy = F.accuracy(data, labels)
    accuracy.backward(None)

    assert data.grad is None
    assert labels.grad is None



import numpy as np

from length.function import Function


class Sigmoid(Function):
    name = "Sigmoid"

    def __init__(self):
        super().__init__()
        # TODO: add more initialization if necessary

    def internal_forward(self, inputs):
        x, = inputs
        # TODO: calculate and return result of sigmoid function
        output = x
        return output,

    def internal_backward(self, inputs, gradients):
        x, = inputs
        grad_in, = gradients
        # TODO: calculate the gradients of this function with respect to its inputs
        grad_x = grad_in
        assert grad_x.shape == x.shape
        return grad_x,


def sigmoid(x):
    """
    This function computes the element-wise sigmoid activation function (https://en.wikipedia.org/wiki/Sigmoid_function)
    on a given input x.
    :param x: the vector, where the activation function shall be applied to
    :return: a vector with the sigmoid activation applied
    """
    return Sigmoid()(x)



import numpy as np

from length import constants
from length.graph import Graph
from length.layers import FullyConnected
from length.tests import gradient_checker
from length.tests.utils import init, retry


def fixed_case():
    data = init([
        [0.44,  0.06, 0.33,  0.76],
        [0.53,  0.65, 0.06, -0.35],
        [0.29, -0.90, 0.86,  0.76],
    ])

    weights = init([
        [-0.74, -0.44],
        [-0.51,  0.63],
        [ 0.73, -0.38],
        [ 0.24, -0.43],
    ])

    bias = init([0.86, 0.63])

    expected = init([
        [0.9271,  0.022],
        [0.0961,  0.934],
        [1.9146, -0.7182],
    ])

    return data, weights, bias, expected


def test_fully_connected_forward():
    data, weights, bias, expected = fixed_case()

    layer = FullyConnected(4, 2)
    layer.weights = weights
    layer.bias = bias

    layer_output = layer(Graph(data))
    gradient_checker.assert_allclose(layer_output.data, expected)


@retry(3)
def test_fully_connected_backward():
    data = np.random.uniform(-1, 1, (10, 50)).astype(constants.DTYPE)

    gradient = np.full((10, 20), 2, dtype=constants.DTYPE)

    layer = FullyConnected(50, 20)
    comp_grad_x, comp_grad_weight, comp_grad_bias = layer.internal_backward((data,), (gradient,))

    f = lambda: layer.internal_forward((data,))
    num_grad_x, num_grad_weight, num_grad_bias = gradient_checker.compute_numerical_gradient(f, (data, layer._weights, layer.bias), (gradient,), eps=1e-2)

    gradient_checker.assert_allclose(comp_grad_x, num_grad_x, atol=1e-4)
    gradient_checker.assert_allclose(comp_grad_weight, num_grad_weight, atol=1e-4)
    gradient_checker.assert_allclose(comp_grad_bias, num_grad_bias, atol=1e-4)


def test_initialization():
    layer = FullyConnected(50, 42)
    assert not (layer.weights == 0).all()



import numpy as np

from length.layer import Layer
from length.optimizers import Adam


def test_adam():
    learning_rate = 0.001
    optimizer = Adam(learning_rate)

    gradients = np.random.random((10, 400))

    layer = Layer()

    deltas, = optimizer.run_update_rule((gradients,), layer)
    np.testing.assert_array_less(deltas, np.full_like(gradients, learning_rate))


def test_adam_increasing_lr():
    learning_rate = 0.001
    optimizer = Adam(learning_rate)

    gradients = np.random.random((10, 400))

    layer = Layer()

    # "collect" momentum (all gradients are positive)
    for i in range(0, 5):
        _, = optimizer.run_update_rule((gradients,), layer)

    inverse_gradients = -1 * gradients

    # do one iteration with negative gradients
    current_deltas, = optimizer.run_update_rule((inverse_gradients,), layer)

    # negative gradients should not immediately reverse the momentum
    np.testing.assert_array_less(np.zeros_like(current_deltas), current_deltas)

    for i in range(0, 5):
        current_deltas, = optimizer.run_update_rule((inverse_gradients,), layer)

    # however multiple iterations should result in negative deltas
    np.testing.assert_array_less(current_deltas, np.zeros_like(current_deltas))



import numpy as np

from length import constants


class Graph:
    """
    Graph stores data and computational history of the neural network
    """
    def __init__(self, data, predecessors=None, creator=None, name=None):
        self.predecessors = predecessors
        self.data = data
        self.creator = creator
        self.grad = None
        self.name = name

    def backward(self, optimizer):
        if self.creator is None:
            # early exit if we are at the top of the computational graph
            return

        # if the size of the data array is 1, we are at the bottom of the computational graph
        # so, we are starting with a gradient of 1
        if self.data.size == 1 and self.grad is None:
            self.grad = np.ones((1,), dtype=constants.DTYPE)

        candidate_layers = []
        seen_layers = set()

        def add_candidate_layer(candidate):
            if candidate is not None and candidate not in seen_layers:
                candidate_layers.append(candidate)
                seen_layers.add(candidate)

        add_candidate_layer(self)

        while candidate_layers:
            candidate_layer = candidate_layers.pop()
            if candidate_layer.creator is None:
                continue

            if candidate_layer.creator.needs_optimizer:
                candidate_layer.creator.optimizer = optimizer

            gradients = candidate_layer.creator.backward(candidate_layer.grad)

            for predecessor, gradient in zip(candidate_layer.predecessors, gradients):
                predecessor.grad = gradient
                if gradient is not None:
                    # the gradient flows to another layer (does not happen with loss layers)
                    add_candidate_layer(predecessor)

    @property
    def shape(self):
        return self.data.shape

    @property
    def dtype(self):
        return self.data.dtype

    def __repr__(self):
        name = getattr(self.creator, 'name', 'input') if self.name is None else self.name
        return "{} {}".format(name, self.shape)

    def visualize(self):
        print(self._visualization_as_str())

    def _visualization_as_str(self):
        layers = [self]

        i = 0
        prev_id = {self: -1}
        while i < len(layers):
            layer = layers[i]
            i += 1
            if layer.creator is None:
                continue

            for predecessor in layer.predecessors:
                if predecessor not in layers:
                    layers.append(predecessor)
                    prev_id[predecessor] = i - 1

        table = [(i - j, repr(layer), i - prev_id[layer]) for j, layer in enumerate(layers)]
        # append header last, since we print in reversed order
        table.append(("id", "layer", "next"))

        widths = [str(max(len(str(row[column])) for row in table)) for column in range(3)]
        template = "{:>" + widths[0] + "} | {:<" + widths[1] + "} | {:" + widths[2] + "}"

        lines = [template.format(*row) for row in table]
        return "\n".join(reversed(lines))



import pytest

from length.function import Function
from length.functions import *
from length.graph import Graph
from length.layer import Layer


def test_function_no_optimizer_necessary():
    # test that a function always says that it does not need an optimizer
    # in order to do that we dynamically get all subclasses of the class function and make sure to not
    # have any subclass of the class layer
    function_subclasses = set(Function.__subclasses__())
    layer_subclasses = set([Layer] + Layer.__subclasses__())
    function_subclasses = function_subclasses.difference(layer_subclasses)

    # we then create an instance of each class and check that it says it wants no optimizer
    for sub_class in function_subclasses:
        assert sub_class.needs_optimizer is False


def test_correct_wrong_input():
    with pytest.raises(AssertionError) as info:
        f = Function()
        f.forward(15)
    message = str(info.value)
    assert "must be a list/tuple" in message
    assert "only includes Graph objects" in message


def test_correct_wrong_input_2():
    with pytest.raises(AssertionError) as info:
        f = Function()
        f.forward([15])
    message = str(info.value)
    assert "must be a list/tuple" in message
    assert "only includes Graph objects" in message


def test_correct_input():
    # this should only raise a not implemented error as internal_forward is not implemented
    with pytest.raises(NotImplementedError):
        f = Function()
        f.forward([Graph(15)])



import numpy as np

from length import constants
from length.functions import relu
from length.functions.relu import Relu
from length.graph import Graph
from length.tests import gradient_checker
from length.tests.utils import init, retry


def test_relu_forward():
    data = init([
        [-0.620304  , -0.1285682 ,  0.4867715 ,  0.09824127],
        [-0.37919873, -0.9272095 , -0.0704312 ,  0.35593647],
        [ 0.19380952,  0.06425636,  0.21729442, -0.3168534 ],
        [-0.62586236, -0.4846    ,  0.84347826,  0.22025743],
        [ 0.02966821, -0.2127131 , -0.33760294, -0.9477733 ]
    ])

    desired = init([
        [0.        , 0.        , 0.4867715 , 0.09824127],
        [0.        , 0.        , 0.        , 0.35593647],
        [0.19380952, 0.06425636, 0.21729442, 0.        ],
        [0.        , 0.        , 0.84347826, 0.22025743],
        [0.02966821, 0.        , 0.        , 0.        ]
    ])

    relu_output = relu(Graph(data))
    np.testing.assert_allclose(relu_output.data, desired)


@retry(3)
def test_relu_backward():
    data = np.random.uniform(-1, 1, (5, 4)).astype(constants.DTYPE)
    gradient = np.random.random(data.shape).astype(dtype=constants.DTYPE)

    data_graph = Graph(data)
    relu_function = Relu()
    relu_function(data_graph)
    computed_gradients, = relu_function.backward(gradient)

    f = lambda: relu_function.internal_forward((data,))
    numerical_gradients, = gradient_checker.compute_numerical_gradient(f, (data,), (gradient,))

    gradient_checker.assert_allclose(computed_gradients, numerical_gradients)



from length.function import Function


class Add(Function):
    name = "Add"

    def internal_forward(self, inputs):
        x, y = inputs
        return x + y

    def internal_backward(self, inputs, gradients):
        gradient, = gradients
        return gradient, gradient


def add(x, y):
    """
    This function performs element-wise addition of two input vectors. Please Note: Both vectors must have the same
    shape!
    :param x: the first vector where each element shall be summed with the elements of the second vector
    :param y: the second vector
    :return: one vector containing the element-wise sum of x and y
    """
    return Add()(x, y)



import setuptools

from pathlib import Path

long_description = Path('README.md').open().readlines()
requirements = [requirement.strip() for requirement in Path('requirements.txt').open().readlines()]

setuptools.setup(
    name="length_hpi",
    version="0.0.1",
    author="Christian Bartz, Joseph Bethe",
    author_email="christian.bartz@hpi.de",
    description="A small framework for teaching students about basics of neural networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    url="https://github.com/HPI-DeepLearning/length",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Education",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.7",
)


import numpy as np

import length.functions as F

from length import constants
from length.graph import Graph
from length.layers import FullyConnected
from length.optimizers.sgd import SGD
from length.tests import gradient_checker


def test_graph_backward_no_layers():
    data = np.array([2], dtype=constants.DTYPE)
    data_graph = Graph(data)
    data_graph.backward(None)
    assert data_graph.grad is None


def test_graph_backward_only_functions_in_graph():
    data = np.array([2], dtype=constants.DTYPE)

    data_graph_1 = Graph(data)
    data_graph_2 = Graph(data)
    h = F.add(data_graph_1, Graph(data))
    h = F.add(h, data_graph_2)

    assert int(h.data) == 6

    h.backward(None)

    assert data_graph_1.grad == 1
    assert data_graph_2.grad == 1
    assert h.grad == 1

    h.grad = 2

    h.backward(None)
    assert data_graph_1.grad == 2
    assert data_graph_2.grad == 2


def test_graph_backward_with_layers():
    # use a fully connected layer and have a look whether the backward pass distributes the gradients correctly
    data = np.random.uniform(-1, 1, (2, 2)).astype(constants.DTYPE)
    labels = np.array([1, 1], dtype=np.int32)

    fc_layer = FullyConnected(2, 2)
    fc_layer.weights[...] = np.zeros_like(fc_layer.weights)
    fc_layer.bias[...] = np.array([-10, 10])

    def run_forward(inputs, labels):
        fc_result = fc_layer(inputs)
        loss = F.softmax_cross_entropy(fc_result, labels)
        return loss

    data_graph = Graph(data)
    label_graph = Graph(labels)

    loss = run_forward(data_graph, label_graph)

    optimizer = SGD(0.001)
    loss.backward(optimizer)

    assert label_graph.grad is None
    gradient_checker.assert_allclose(data_graph.grad, np.zeros_like(data_graph.grad))
    gradient_checker.assert_allclose(fc_layer.weights, np.zeros_like(fc_layer.weights))
    gradient_checker.assert_allclose(fc_layer.bias, np.array([-10, 10]))

    # change the labels and make sure that the gradients are different now
    # the absolute values of the gradients of one sample shall be higher than the gradients of the other sample
    labels = np.array([0, 1], dtype=np.int32)
    label_graph = Graph(labels)

    loss = run_forward(data_graph, label_graph)

    loss.backward(optimizer)

    assert label_graph.grad is None
    assert (np.abs(data_graph.grad[0]) > np.abs(data_graph.grad[1])).all()



import numpy as np

from length import constants
from length.functions import sigmoid
from length.functions.sigmoid import Sigmoid
from length.graph import Graph
from length.tests import gradient_checker
from length.tests.utils import init, retry


def test_sigmoid_forward():
    data = init([
        [-0.22342056,  0.6927312 ],
        [ 0.4227562 , -0.59764487],
        [ 0.7870561 ,  0.372502  ]
    ])

    sigmoid_output = sigmoid(Graph(data))
    desired = init([
        [0.44437608, 0.66657424],
        [0.6041426 , 0.3548827 ],
        [0.6871989 , 0.5920634 ]
    ])

    np.testing.assert_allclose(sigmoid_output.data, desired)


@retry(3)
def test_sigmoid_backward():
    data = np.random.uniform(-1, 1, (3, 2)).astype(constants.DTYPE)
    gradient = np.random.random(data.shape).astype(constants.DTYPE)

    data_graph = Graph(data)
    sigmoid_function = Sigmoid()
    sigmoid_function(data_graph)
    computed_gradients, = sigmoid_function.backward(gradient)

    f = lambda: sigmoid_function.internal_forward((data,))
    numerical_gradients, = gradient_checker.compute_numerical_gradient(f, (data,), (gradient,))

    gradient_checker.assert_allclose(computed_gradients, numerical_gradients, atol=1e-4, rtol=1e-3)



import numpy as np


def compute_numerical_gradient(function, inputs, last_gradients, eps=1e-3):
    """
    Function that computes numerical gradient (difference quotient explanation:https://en.wikipedia.org/wiki/Difference_quotient)
    for a given input. This function is to be used to check the backward implementation of each function/layer
    :param function: a python function that takes no arguments and performs a forward pass of the function/layer in question
    :param inputs: a tuple of input arrays that is used to perform the forward pass of the function/layer
    :param last_gradients: the gradients that are to be passed to the function/layer during backward
    :param eps: epsilon to be used for calculation of numerical gradient
    :return: the numerical gradient of the function
    """
    gradients = tuple(np.zeros_like(x) for x in inputs)

    for x, gx in zip(inputs, gradients):
        flat_x = x.reshape(-1)
        flat_gx = gx.reshape(-1)

        for i in range(flat_x.size):
            original_value = flat_x[i]
            flat_x[i] = original_value - eps
            outputs_1 = function()
            flat_x[i] = original_value + eps
            outputs_2 = function()
            flat_x[i] = original_value

            for y1, y2, gy in zip(outputs_1, outputs_2, last_gradients):
                if gy is not None:
                    # calculate the difference quotient
                    numerator = float(sum(((y2 - y1) * gy).reshape(-1)))
                    denominator = 2 * eps
                    grad = numerator / denominator
                    flat_gx[i] += grad

    return gradients


def assert_allclose(actual, desired, atol=1e-5, rtol=1e-4, verbose=True):
    """
    Wrapper around numpy allclose assertion that uses better tolerance values for neural network computation
    :param actual: the first value to check
    :param desired: the second value to check
    :param atol: absolute allowed tolerance of values
    :param rtol: realtive allowed tolerance of values
    :param verbose: verbose output of allclose check
    """
    np.testing.assert_allclose(actual, desired, rtol=rtol, atol=atol, verbose=verbose)



from length.graph import Graph


class Function:
    """
    Function is the base class for all computations of the neural network
    a function does not have any parameters that need to be optimized
    """
    needs_optimizer = False
    name = "Function"

    def __init__(self):
        self.inputs = None
        self.outputs = None

    def internal_forward(self, inputs):
        """
        inputs is a tuple of numpy arrays
        :returns outputs as tuple of numpy arrays
        """
        raise NotImplementedError

    def internal_backward(self, inputs, gradients):
        """
        :param inputs: is a tuple of numpy arrays
        :param gradients: is a tuple of numpy arrays
        :returns gradients as tuple of numpy arrays (first elements are gradients with respect to the inputs)
        """
        raise NotImplementedError

    def forward(self, graphs):
        requirement = "The input to forward must be a list/tuple which only includes Graph objects."
        assert isinstance(graphs, (list, tuple)), requirement
        assert all(isinstance(graph, Graph) for graph in graphs), requirement

        self.inputs = tuple(graph.data for graph in graphs)
        self.outputs = self.internal_forward(self.inputs)

        output_graphs = [Graph(output, predecessors=graphs, creator=self) for output in self.outputs]
        if len(output_graphs) == 1:
            return output_graphs[0]
        return output_graphs

    def backward(self, gradients):
        gradients = self.internal_backward(self.inputs, (gradients,))
        return gradients

    def __call__(self, *args):
        return self.forward(args)





class Initializer:
    """
    Base structure for all initializers
    """
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, array):
        """
        initialize the given array according to the initialization rule implemented in this initializer
        :param array: the array to initialize
        :return: the initialized array
        """
        raise NotImplementedError
